__all__ = ("DeepDrillingSurvey", "generate_dd_surveys", "dd_bfs")

import copy
import logging
import random
from functools import cached_property

import numpy as np

import rubin_scheduler.scheduler.basis_functions as basis_functions
from rubin_scheduler.scheduler import features
from rubin_scheduler.scheduler.surveys import BaseSurvey
from rubin_scheduler.scheduler.utils import ObservationArray
from rubin_scheduler.utils import DEFAULT_NSIDE, ddf_locations, ra_dec2_hpid

log = logging.getLogger(__name__)


class DeepDrillingSurvey(BaseSurvey):
    """A survey class for running deep drilling fields.

    Parameters
    ----------
    basis_functions : list of rubin_scheduler.scheduler.basis_function
        These should be feasibility basis functions.
    RA : float
        The RA of the field (degrees)
    dec : float
        The dec of the field to observe (degrees)
    sequence : list of observation objects or str (rgizy)
        The sequence of observations to take. Can be a string of
        list of obs objects.
    nvis : list of ints
        The number of visits in each band. Should be same length
        as sequence.
    survey_name : str (DD)
        The name to give this survey so it can be tracked
    reward_value : float (101.)
        The reward value to report if it is able to start (unitless).
    readtime : float (2.)
        Readout time for computing approximate time of observing
        the sequence. (seconds)
    flush_pad : float (30.)
        How long to hold observations in the queue after they
        were expected to be completed (minutes).
    """

    def __init__(
        self,
        basis_functions,
        RA,
        dec,
        sequence="rgizy",
        nvis=[20, 10, 20, 26, 20],
        exptime=30.0,
        u_exptime=30.0,
        nexp=2,
        ignore_obs=None,
        survey_name="DD",
        reward_value=None,
        readtime=2.0,
        band_change_time=120.0,
        nside=DEFAULT_NSIDE,
        flush_pad=30.0,
        seed=42,
        detailers=None,
        science_program=None,
    ):
        super(DeepDrillingSurvey, self).__init__(
            nside=nside,
            basis_functions=basis_functions,
            detailers=detailers,
            ignore_obs=ignore_obs,
            science_program=science_program,
        )
        random.seed(a=seed)

        self.ra = np.radians(RA)
        self.ra_hours = RA / 360.0 * 24.0
        self.dec = np.radians(dec)
        self.survey_name = survey_name
        self.reward_value = reward_value
        self.flush_pad = flush_pad / 60.0 / 24.0  # To days
        self.band_sequence = []
        if isinstance(sequence, str):
            self.observations = []
            for num, bandname in zip(nvis, sequence):
                for j in range(num):
                    obs = ObservationArray()
                    obs["band"] = bandname
                    if bandname == "u":
                        obs["exptime"] = u_exptime
                    else:
                        obs["exptime"] = exptime
                    obs["RA"] = self.ra
                    obs["dec"] = self.dec
                    obs["nexp"] = nexp
                    obs["scheduler_note"] = survey_name
                    self.observations.append(obs)
        else:
            self.observations = sequence

        # Let's just make this an array for ease of use
        self.observations = np.concatenate(self.observations)
        order = np.argsort(self.observations["band"])
        self.observations = self.observations[order]

        n_band_change = np.size(np.unique(self.observations["band"]))

        # Make an estimate of how long a seqeunce will take.
        # Assumes no major rotational or spatial
        # dithering slowing things down.
        self.approx_time = (
            np.sum(self.observations["exptime"] + readtime * self.observations["nexp"]) / 3600.0 / 24.0
            + band_change_time * n_band_change / 3600.0 / 24.0
        )  # to days

        if self.reward_value is None:
            self.extra_features["Ntot"] = features.NObsCount()
            self.extra_features["N_survey"] = features.NObsCount(note=self.survey_name)

    @cached_property
    def roi_hpid(self):
        hpid = ra_dec2_hpid(self.nside, np.degrees(self.ra), np.degrees(self.dec))
        return hpid

    def check_continue(self, observation, conditions):
        # feasibility basis functions?
        """
        This method enables external calls to check if a given
        observations that belongs to this survey is
        feasible or not. This is called once a sequence has
        started to make sure it can continue.

        XXX--TODO:  Need to decide if we want to develope check_continue,
        or instead hold the sequence in the survey, and be able to check
        it that way.
        """

        result = True

        return result

    def calc_reward_function(self, conditions):
        result = -np.inf
        if self._check_feasibility(conditions):
            if self.reward_value is not None:
                result = self.reward_value
            else:
                # XXX This might backfire if we want to have DDFs with
                # different fractions of the survey time. Then might need
                # to define a goal fraction, and have the reward be the
                # number of observations behind that target fraction.
                result = self.extra_features["Ntot"].feature / (self.extra_features["N_survey"].feature + 1)
        return result

    def generate_observations_rough(self, conditions):
        result = []
        if self._check_feasibility(conditions):
            result = copy.deepcopy(self.observations)

            # Set the flush_by
            result["flush_by_mjd"] = conditions.mjd + self.approx_time + self.flush_pad

            # remove bands that are not mounted
            mask = np.isin(result["band"], conditions.mounted_bands)
            result = result[mask]
            # Put current loaded band first
            ind1 = np.where(result["band"] == conditions.current_band)[0]
            ind2 = np.where(result["band"] != conditions.current_band)[0]
            result = result[ind1.tolist() + (ind2.tolist())]

        return result

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} survey_name='{self.survey_name}'"
            f", RA={self.ra}, dec={self.dec} at {hex(id(self))}>"
        )


def dd_bfs(
    RA,
    dec,
    survey_name,
    ha_limits,
    frac_total=0.0185 / 2.0,
    aggressive_frac=0.011 / 2.0,
    delays=[0.0, 0.5, 1.5],
    time_needed=62.0,
    wind_speed_maximum=20.0,
    nside=None,
):
    """
    Convienence function to generate all the feasibility basis functions
    """
    sun_alt_limit = -18.0
    fractions = [0.00, aggressive_frac, frac_total]
    bfs = []
    bfs.append(basis_functions.NotTwilightBasisFunction(sun_alt_limit=sun_alt_limit))
    bfs.append(basis_functions.TimeToTwilightBasisFunction(time_needed=time_needed))
    bfs.append(basis_functions.AvoidDirectWind(wind_speed_maximum=wind_speed_maximum, nside=nside))
    bfs.append(basis_functions.HourAngleLimitBasisFunction(RA=RA, ha_limits=ha_limits))
    bfs.append(basis_functions.MoonDownBasisFunction())
    bfs.append(basis_functions.FractionOfObsBasisFunction(frac_total=frac_total, survey_name=survey_name))
    bfs.append(
        basis_functions.LookAheadDdfBasisFunction(
            frac_total,
            aggressive_frac,
            sun_alt_limit=sun_alt_limit,
            time_needed=time_needed,
            RA=RA,
            survey_name=survey_name,
            ha_limits=ha_limits,
        )
    )
    bfs.append(
        basis_functions.SoftDelayBasisFunction(fractions=fractions, delays=delays, survey_name=survey_name)
    )
    bfs.append(basis_functions.TimeToScheduledBasisFunction(time_needed=time_needed))

    return bfs


def generate_dd_surveys(
    nside=None,
    nexp=2,
    detailers=None,
    euclid_detailers=None,
    reward_value=100,
    frac_total=0.0185 / 2.0,
    aggressive_frac=0.011 / 2.0,
    exptime=30,
    u_exptime=30,
    nvis_master=[8, 20, 10, 20, 26, 20],
    delays=[0.0, 0.5, 1.5],
):
    """Utility to return a list of standard deep drilling field surveys.

    XXX-Someone double check that I got the coordinates right!

    """

    if euclid_detailers is None:
        euclid_detailers = detailers

    surveys = []

    locations = ddf_locations()

    # ELAIS S1
    survey_name = "DD:ELAISS1"
    RA = locations["ELAISS1"][0]
    dec = locations["ELAISS1"][1]
    ha_limits = ([0.0, 1.5], [21.5, 24.0])
    bfs = dd_bfs(
        RA,
        dec,
        survey_name,
        ha_limits,
        frac_total=frac_total,
        aggressive_frac=aggressive_frac,
        delays=delays,
        nside=nside,
    )
    surveys.append(
        DeepDrillingSurvey(
            bfs,
            RA,
            dec,
            sequence="urgizy",
            nvis=nvis_master,
            exptime=exptime,
            u_exptime=u_exptime,
            survey_name=survey_name,
            reward_value=reward_value,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # XMM-LSS
    survey_name = "DD:XMM-LSS"
    RA = locations["XMM_LSS"][0]
    dec = locations["XMM_LSS"][1]
    ha_limits = ([0.0, 1.5], [21.5, 24.0])
    bfs = dd_bfs(
        RA,
        dec,
        survey_name,
        ha_limits,
        frac_total=frac_total,
        aggressive_frac=aggressive_frac,
        delays=delays,
        nside=nside,
    )

    surveys.append(
        DeepDrillingSurvey(
            bfs,
            RA,
            dec,
            sequence="urgizy",
            exptime=exptime,
            u_exptime=u_exptime,
            nvis=nvis_master,
            survey_name=survey_name,
            reward_value=reward_value,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # Extended Chandra Deep Field South
    survey_name = "DD:ECDFS"
    RA = locations["ECDFS"][0]
    dec = locations["ECDFS"][1]
    ha_limits = [[0.5, 3.0], [20.0, 22.5]]
    bfs = dd_bfs(
        RA,
        dec,
        survey_name,
        ha_limits,
        frac_total=frac_total,
        aggressive_frac=aggressive_frac,
        delays=delays,
        nside=nside,
    )
    surveys.append(
        DeepDrillingSurvey(
            bfs,
            RA,
            dec,
            sequence="urgizy",
            nvis=nvis_master,
            exptime=exptime,
            u_exptime=u_exptime,
            survey_name=survey_name,
            reward_value=reward_value,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # COSMOS
    survey_name = "DD:COSMOS"
    RA = locations["COSMOS"][0]
    dec = locations["COSMOS"][1]
    ha_limits = ([0.0, 2.5], [21.5, 24.0])
    bfs = dd_bfs(
        RA,
        dec,
        survey_name,
        ha_limits,
        frac_total=frac_total,
        aggressive_frac=aggressive_frac,
        delays=delays,
        nside=nside,
    )
    surveys.append(
        DeepDrillingSurvey(
            bfs,
            RA,
            dec,
            sequence="urgizy",
            nvis=nvis_master,
            exptime=exptime,
            u_exptime=u_exptime,
            survey_name=survey_name,
            reward_value=reward_value,
            nside=nside,
            nexp=nexp,
            detailers=detailers,
        )
    )

    # Euclid Fields
    # I can use the sequence kwarg to do two positions per sequence
    bands = "urgizy"
    nviss = nvis_master
    survey_name = "DD:EDFS"
    # Note the sequences need to be in radians since they are using
    # observation objects directly
    # Coords from jc.cuillandre@cea.fr Oct 15, 2020
    r_as = np.radians([locations["EDFS_a"][0], locations["EDFS_b"][0]])
    decs = np.radians([locations["EDFS_a"][1], locations["EDFS_b"][1]])
    suffixes = [", a", ", b"]
    sequence = []

    for bandname, nvis in zip(bands, nviss):
        for ra, dec, suffix in zip(r_as, decs, suffixes):
            for num in range(nvis):
                obs = ObservationArray()
                obs["band"] = bandname
                if bandname == "u":
                    obs["exptime"] = u_exptime
                else:
                    obs["exptime"] = exptime
                obs["RA"] = ra
                obs["dec"] = dec
                obs["nexp"] = nexp
                obs["scheduler_note"] = survey_name + suffix
                sequence.append(obs)

    ha_limits = ([0.0, 1.5], [22.5, 24.0])
    # And back to degrees for the basis function. Need to bump up the
    # time needed since it's a double field.
    bfs = dd_bfs(
        np.degrees(r_as[0]),
        np.degrees(decs[0]),
        survey_name,
        ha_limits,
        frac_total=frac_total,
        aggressive_frac=aggressive_frac,
        delays=delays,
        time_needed=120.0,
    )
    surveys.append(
        DeepDrillingSurvey(
            bfs,
            np.degrees(r_as),
            np.degrees(decs),
            sequence=sequence,
            survey_name=survey_name,
            reward_value=reward_value,
            nside=nside,
            nexp=nexp,
            detailers=euclid_detailers,
        )
    )

    return surveys



import unittest

from rubin_scheduler.scheduler.surveys import ToOScriptedSurvey


class TestToO(unittest.TestCase):

    def test_too(self):
        survey = ToOScriptedSurvey([])
        ra1, dec1 = survey._tesselate([100, 101])
        # This should spin things
        ra2, dec2 = survey._tesselate([100, 101])

        assert ra2[0] != ra1[0]

        # If we ask for a single HEALpix
        # Should only give a pointing at
        # one location

        ra1, dec1 = survey._tesselate([100])

        assert len(ra1) == 1

        # Should be no change if we call again
        ra2, dec2 = survey._tesselate([100])

        assert ra1 == ra2


if __name__ == "__main__":
    unittest.main()



import os
import unittest

import numpy as np

import rubin_scheduler.utils as utils
from rubin_scheduler.data import get_data_dir
from rubin_scheduler.scheduler import sim_runner
from rubin_scheduler.scheduler.basis_functions import SunAltLimitBasisFunction
from rubin_scheduler.scheduler.example import simple_greedy_survey, simple_pairs_survey
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.schedulers import CoreScheduler

SAMPLE_BIG_DATA_FILE = os.path.join(get_data_dir(), "scheduler/dust_maps/dust_nside_32.npz")


class TestNside(unittest.TestCase):
    @unittest.skipUnless(os.path.isfile(SAMPLE_BIG_DATA_FILE), "Test data not available.")
    def test_nside(self):
        """
        test running at higher nside
        """
        mjd_start = utils.SURVEY_START_MJD
        nside = 64
        observatory = ModelObservatory(
            nside=nside, mjd_start=mjd_start, cloud_data="ideal", downtimes="ideal"
        )
        conditions = observatory.return_conditions()
        # set the mjd to start of night
        observatory.mjd = conditions.sun_n12_setting + 1.01
        survey_length = 0.1  # days

        # Add an avoidance of twilight+ for the pairs surveys -
        # this ensures greedy survey will have some time to operate
        pairs_surveys = [
            simple_pairs_survey(bandname="g", bandname2="r", nside=nside),
            simple_pairs_survey(bandname="i", bandname2="z", nside=nside),
        ]
        for survey in pairs_surveys:
            survey.basis_functions.append(SunAltLimitBasisFunction(alt_limit=-22))
            survey.basis_weights.append(0)
        greedy_surveys = [
            simple_greedy_survey(bandname="z", nside=nside),
        ]

        scheduler = CoreScheduler([pairs_surveys, greedy_surveys], nside=nside)

        observatory, scheduler, observations = sim_runner(
            observatory, scheduler, sim_duration=survey_length, filename=None
        )

        # Make sure some blobs executed
        assert "simple pair 30, iz, a" in observations["scheduler_note"]
        assert "simple pair 30, iz, b" in observations["scheduler_note"]
        # Make sure some greedy executed
        assert "simple greedy z" in observations["scheduler_note"]
        # Make sure lots of observations executed
        assert observations.size > 50
        # Make sure nothing tried to look through the earth
        assert np.min(observations["alt"]) > 0


if __name__ == "__main__":
    unittest.main()



__all__ = ("bearing", "dest_latlon", "point_to_line_distance")

import numpy as np

from .coordinate_transformations import _angular_separation


def bearing(lon1, lat1, lon2, lat2):
    """Bearing between two points
    all radians
    """

    delta_l = lon2 - lon1
    X = np.cos(lat2) * np.sin(delta_l)
    Y = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(delta_l)
    theta = np.arctan2(X, Y)

    return theta


def dest_latlon(dist, bearing, lat1, lon1):
    """Destination lat and lon given a distance, bearing and starting position
    all radians
    """

    lat2 = np.arcsin(np.sin(lat1) * np.cos(dist) + np.cos(lat1) * np.sin(dist) * np.cos(bearing))
    lon2 = lon1 + np.arctan2(
        np.sin(bearing) * np.sin(dist) * np.cos(lat1),
        np.cos(dist) - np.sin(lat1) * np.sin(lat2),
    )
    return lat2, lon2


def point_to_line_distance(lon1, lat1, lon2, lat2, lon3, lat3):
    """All radians
    points 1 and 2 define an arc segment,
    this finds the distance of point 3 to the arc segment.
    """

    result = lon1 * 0
    needed = np.ones(result.size, dtype=bool)

    bear12 = bearing(lon1, lat1, lon2, lat2)
    bear13 = bearing(lon1, lat1, lon3, lat3)
    dis13 = _angular_separation(lon1, lat1, lon3, lat3)

    # Is relative bearing obtuse?
    diff = np.abs(bear13 - bear12)
    if np.size(diff) == 1:
        if diff > np.pi:
            diff = 2 * np.pi - diff
        if diff > (np.pi / 2):
            return dis13
    else:
        over = np.where(diff > np.pi)
        diff[over] = 2 * np.pi - diff[over]
        solved = np.where(diff > (np.pi / 2))[0]
        result[solved] = dis13[solved]
        needed[solved] = False

    # Find the cross-track distance.
    dxt = np.arcsin(np.sin(dis13) * np.sin(bear13 - bear12))

    # Is p4 beyond the arc?
    dis12 = _angular_separation(lon1, lat1, lon2, lat2)
    dis14 = np.arccos(np.cos(dis13) / np.cos(dxt))
    if np.size(dis14) == 1:
        if dis14 > dis12:
            return _angular_separation(lon2, lat2, lon3, lat3)
    else:
        solved = np.where((dis14 > dis12) & needed)[0]
        result[solved] = _angular_separation(lon2[solved], lat2[solved], lon3[solved], lat3[solved])
        needed[solved] = False

    if np.size(lon1) == 1:
        return np.abs(dxt)
    else:
        result[needed] = np.abs(dxt[needed])
        return result



import unittest

import healpy as hp
import numpy as np

import rubin_scheduler.utils as utils


class TestHealUtils(unittest.TestCase):
    def test_ra_decs_rad(self):
        """
        Test that the Ra Dec conversions round-trip
        """

        nside = 64
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils._hpid2_ra_dec(nside, hpids)

        hpids_return = utils._ra_dec2_hpid(nside, ra, dec)

        np.testing.assert_array_equal(hpids, hpids_return)

    def test_ra_decs_deg(self):
        """
        Test that the Ra Dec conversions round-trip
        """

        nside = 64
        hpids = np.arange(hp.nside2npix(nside))
        ra, dec = utils.hpid2_ra_dec(nside, hpids)

        hpids_return = utils.ra_dec2_hpid(nside, ra, dec)

        np.testing.assert_array_equal(hpids, hpids_return)

    def test_bin_rad(self):
        """
        Test that healbin returns correct values and valid healpy maps.
        """

        ra = np.zeros(3)
        dec = np.zeros(3)
        values = ra * 0.0 + 1.0

        nside = 128
        hpid = utils._ra_dec2_hpid(nside, ra[0], dec[0])

        map1 = utils._healbin(ra, dec, values, nside=nside)
        self.assertEqual(map1[hpid], 1.0)
        self.assertEqual(hp.maptype(map1), 0)
        map2 = utils._healbin(ra, dec, values, nside=nside, reduce_func=np.sum)
        self.assertEqual(map2[hpid], 3.0)
        self.assertEqual(hp.maptype(map2), 0)
        map3 = utils._healbin(ra, dec, values, nside=nside, reduce_func=np.std)
        self.assertEqual(map3[hpid], 0.0)
        self.assertEqual(hp.maptype(map3), 0)

    def test_bin_deg(self):
        """
        Test that healbin returns correct values and valid healpy maps.
        """

        ra = np.zeros(3)
        dec = np.zeros(3)
        values = ra * 0.0 + 1.0

        nside = 128
        hpid = utils.ra_dec2_hpid(nside, ra[0], dec[0])

        map1 = utils.healbin(ra, dec, values, nside=nside)
        self.assertEqual(map1[hpid], 1.0)
        self.assertEqual(hp.maptype(map1), 0)
        map2 = utils.healbin(ra, dec, values, nside=nside, reduce_func=np.sum)
        self.assertEqual(map2[hpid], 3.0)
        self.assertEqual(hp.maptype(map2), 0)
        map3 = utils.healbin(ra, dec, values, nside=nside, reduce_func=np.std)
        self.assertEqual(map3[hpid], 0.0)
        self.assertEqual(hp.maptype(map3), 0)

    def test_mask_grow(self):
        """Test we can grow a healpix mask map"""

        nside = 32
        nan_indx = tuple([0, 100])
        scale = hp.nside2resol(nside)

        to_mask = utils._hp_grow_mask(nside, nan_indx, grow_size=scale * 2)

        # That should have made some things mask
        assert 100 > np.size(to_mask) > 5

        # Test another nside
        nside = 128
        nan_indx = tuple([0, 100])
        scale = hp.nside2resol(nside)
        to_mask = utils._hp_grow_mask(nside, nan_indx, grow_size=scale * 2)

        # That should have made some things mask
        assert 100 > np.size(to_mask) > 5

        # Do a silly loop check to make sure things got masked properly
        # Need to turn off the machine precision kwarg
        nside = 128
        nan_indx = tuple([0, 100])
        for scale in np.radians(
            [
                0.1,
                1.0,
                10.0,
                30,
            ]
        ):
            to_mask = utils._hp_grow_mask(nside, nan_indx, grow_size=scale, scale=None)

            ra, dec = utils._hpid2_ra_dec(nside, np.arange(hp.nside2npix(nside)))
            to_mask_set = set(to_mask)
            all_close = []
            for indx in nan_indx:
                distances = utils._angular_separation(ra, dec, ra[indx], dec[indx])
                close = np.where(distances <= scale)[0]
                all_close.extend(close)
                assert set(close).issubset(to_mask_set)
            all_close = np.unique(all_close)
            # Make sure we aren't including any extra pixels
            assert np.size(all_close) == np.size(to_mask)

        # Check that 0 distance doesn't mask anything new
        to_mask = utils._hp_grow_mask(nside, nan_indx, grow_size=0)
        assert np.size(to_mask) == np.size(nan_indx)
        # Check scale can be wiped out.
        to_mask = utils._hp_grow_mask(nside, nan_indx, grow_size=0, scale=None)
        assert np.size(to_mask) == np.size(nan_indx)


if __name__ == "__main__":
    unittest.main()



import warnings

import numpy as np

from rubin_scheduler.scheduler.detailers import BaseDetailer


class BandPickToODetailer(BaseDetailer):
    """Select which filters are used based on information in a
    ToO event.

    Parameters
    ----------
    band_start : `str`
        Which band to consider changing
    band_end : `str`
        What band to change to.
    distance_limit : `float`
        For distances less than distance_limit,
        the bands are swapped. (kpc)
    check_mounted : `bool`
        Check that the desired band is mounted before changing
        values.
    require_dark : `bool`
        If in dark conditions, the bands are swapped.
    """

    def __init__(
        self, band_start="i", band_end="z", distance_limit=10, check_mounted=True, require_dark=False
    ):
        self.band_start = band_start
        self.band_end = band_end
        self.distance_limit = distance_limit
        self.check_mounted = check_mounted
        self.require_dark = require_dark

    def __call__(self, observations, conditions, target_o_o=None):

        if hasattr(target_o_o, "posterior_distance"):
            if target_o_o.posterior_distance is not None:
                if self.require_dark and conditions.moon_phase > 50:
                    warnings.warn("Requires dark time, moon phase is greater than 50%.")
                elif target_o_o.posterior_distance < self.distance_limit:
                    if self.check_mounted:
                        mounted = self.band_end in conditions.mounted_bands
                    else:
                        mounted = True
                    if mounted:
                        band_start_indx = np.where(observations["band"] == self.band_start)[0]
                        observations["band"][band_start_indx] = self.band_end
            else:
                warnings.warn("No distance present, not swapping bands.")
        else:
            warnings.warn("No distance present, not swapping bands.")
        return observations



__all__ = ("DEFAULT_NSIDE", "SURVEY_START_MJD", "survey_start_mjd")

import warnings

from astropy.time import Time

DEFAULT_NSIDE = 32  # HEALpix nside, ~1.83 degree resolution
SURVEY_START_MJD = Time("2025-11-01T12:00:00").mjd  # Should be during the day before first observation


def survey_start_mjd():
    """For backwards compatibility"""
    warnings.warn("Function survey_start_mjd is deprecated, use SURVEY_START_MJD.", FutureWarning)

    return SURVEY_START_MJD



import unittest
import warnings

import numpy as np

import rubin_scheduler.scheduler.basis_functions as basis_functions
from rubin_scheduler.scheduler.features import Conditions
from rubin_scheduler.scheduler.model_observatory import ModelObservatory
from rubin_scheduler.scheduler.utils import ObservationArray


class TestBasis(unittest.TestCase):
    def setUp(self) -> None:
        self.observatory = ModelObservatory()
        self.conditions = self.observatory.return_conditions()

    def test_basics(self):
        """Test the basics of each basis function"""

        # List of simple basis functions that don't have
        # any required args
        bfs = [
            basis_functions.ConstantBasisFunction,
            basis_functions.DelayStartBasisFunction,
            basis_functions.AvoidFastRevisitsBasisFunction,
            basis_functions.VisitRepeatBasisFunction,
            basis_functions.M5DiffBasisFunction,
            basis_functions.StrictBandBasisFunction,
            basis_functions.BandChangeBasisFunction,
            basis_functions.SlewtimeBasisFunction,
            basis_functions.AzModuloBasisFunction,
            basis_functions.DecModuloBasisFunction,
            basis_functions.NearSunHighAirmassBasisFunction,
            basis_functions.EclipticBasisFunction,
            basis_functions.NGoodSeeingBasisFunction,
            basis_functions.AvoidDirectWind,
            basis_functions.BandDistBasisFunction,
            basis_functions.RewardRisingBasisFunction,
            basis_functions.BandLoadedBasisFunction,
            basis_functions.OnceInNightBasisFunction,
            basis_functions.SunAltHighLimitBasisFunction,
            basis_functions.TimeToTwilightBasisFunction,
            basis_functions.NotTwilightBasisFunction,
            basis_functions.AfterEveningTwiBasisFunction,
            basis_functions.ForceDelayBasisFunction,
            basis_functions.HourAngleLimitBasisFunction,
            basis_functions.MoonDownBasisFunction,
            basis_functions.CloudedOutBasisFunction,
            basis_functions.SoftDelayBasisFunction,
            basis_functions.SunAltLimitBasisFunction,
            basis_functions.NightModuloBasisFunction,
            basis_functions.EndOfEveningBasisFunction,
            basis_functions.TimeToScheduledBasisFunction,
            basis_functions.CloseToTwilightBasisFunction,
            basis_functions.InTimeWindowBasisFunction,
            basis_functions.HaMaskBasisFunction,
            basis_functions.MoonAvoidanceBasisFunction,
            basis_functions.MapCloudBasisFunction,
            basis_functions.PlanetMaskBasisFunction,
            basis_functions.SolarElongationMaskBasisFunction,
            basis_functions.AltAzShadowMaskBasisFunction,
            basis_functions.AltAzShadowTimeLimitedBasisFunction,
            basis_functions.MoonAltLimitBasisFunction,
            basis_functions.RevHaMaskBasisFunction,
            basis_functions.MaskAllButNES,
            basis_functions.NInNightMaskBasisFunction,
            basis_functions.OnlyBeforeNightBasisFunction,
            basis_functions.MaskAfterNObsBasisFunction,
        ]

        obs = ObservationArray()
        obs["band"] = "r"
        obs["mjd"] = 59000.0
        indx = np.array([1000])

        for bf in bfs:
            awake_bf = bf()
            awake_bf.add_observation(obs, indx=indx)
            feas = awake_bf.check_feasibility(conditions=self.conditions)
            reward = awake_bf(self.conditions)
            assert feas is not None
            assert reward is not None

    def test_altazshadowtimelimited(self):
        cond = Conditions()
        bf = basis_functions.AltAzShadowTimeLimitedBasisFunction()

        # Set conditions to be in the middle of the night,
        # bf should return 0
        cond.mjd = 5900
        # Sunrise is in 4 hours
        cond.sunrise = 5900 + 4.0 / 24.0
        cond.sunset = 5900 - 4.0 / 24.0

        assert bf(cond) == 0

        # set to be close to sunrise/set, should return mask
        cond.sunrise = 5900 + 1.0 / 24.0

        mask1 = bf(cond)
        assert np.size(mask1) > 1

        # Make a more restrictive mask
        bf = basis_functions.AltAzShadowTimeLimitedBasisFunction(min_alt=40, max_alt=50)
        mask2 = bf(cond)

        # New mask should have fewer valid pixels than default values
        assert np.size(mask2) > 2
        assert np.sum(np.isfinite(mask1)) > np.sum(np.isfinite(mask2))

    def test_slewtime_basis_function(self):
        current_band = self.conditions.current_filter
        # Test value with matching bandpass
        bf = basis_functions.SlewtimeBasisFunction(bandname=current_band)
        reward = bf(self.conditions)
        # Slewtime reward is always < 0 ..
        self.assertTrue(np.nanmax(reward) < 0)
        # Test value with different bandpass
        bands = "ugrizy"
        different_band = (bands.index(current_band) + 2) % 5
        different_band = bands[different_band]
        bf = basis_functions.SlewtimeBasisFunction(bandname=different_band)
        reward = bf(self.conditions)
        self.assertTrue(np.nanmax(reward) == 0)
        # Test value with bandpass = None
        bf = basis_functions.SlewtimeBasisFunction(bandname=None)
        reward = bf(self.conditions)
        self.assertTrue(np.nanmax(reward) < 0)

    def test_visit_repeat_basis_function(self):
        bf = basis_functions.VisitRepeatBasisFunction()

        indx = np.array([1000])

        # 30 minute step
        delta = 30.0 / 60.0 / 24.0

        # Add 1st observation, should still be zero
        obs = ObservationArray()
        obs["band"] = "r"
        obs["mjd"] = 59000.0
        conditions = Conditions()
        conditions.mjd = np.max(obs["mjd"])
        bf.add_observation(obs, indx=indx)
        self.assertEqual(np.max(bf(conditions)), 0.0)

        # Advance time so now we want a pair
        conditions.mjd += delta
        self.assertEqual(np.max(bf(conditions)), 1.0)

        # Now complete the pair and it should go back to zero
        bf.add_observation(obs, indx=indx)

        conditions.mjd += delta
        self.assertEqual(np.max(bf(conditions)), 0.0)

    def test_force_delay(self):
        bf = basis_functions.ForceDelayBasisFunction(days_delay=3.0, scheduler_note="survey")
        obs = ObservationArray()
        obs["scheduler_note"] = "not_match"
        obs["mjd"] = 10
        bf.add_observation(obs)
        conditions = Conditions()
        conditions.mjd = 11.0
        assert bf.check_feasibility(conditions)
        # Now it matches, so should block
        obs["scheduler_note"] = "survey"
        bf.add_observation(obs)

        assert not bf.check_feasibility(conditions)

    def test_label(self):
        bf = basis_functions.VisitRepeatBasisFunction()
        self.assertIsInstance(bf.label(), str)

        bf = basis_functions.SlewtimeBasisFunction(nside=16)
        self.assertIsInstance(bf.label(), str)

    def test_visit_gap(self):
        visit_gap = basis_functions.VisitGap(note="test")

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = ObservationArray()
        observation["band"] = "r"
        observation["scheduler_note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["scheduler_note"] = "test"
        visit_gap.add_observation(observation=observation)

        # now observation with the correct note
        assert not visit_gap.check_feasibility(conditions=conditions)

        # check it becomes feasible again once enough time has passed
        conditions.mjd += 2.0 * visit_gap.gap

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_visit_gap_with_band(self):
        visit_gap = basis_functions.VisitGap(note="test", band_names=["g"])

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = ObservationArray()
        observation["band"] = "r"
        observation["scheduler_note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["scheduler_note"] = "test"
        visit_gap.add_observation(observation=observation)

        # observation with the wrong band
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["band"] = "g"
        visit_gap.add_observation(observation=observation)

        # observation with the correct note and band
        assert not visit_gap.check_feasibility(conditions=conditions)

        # check it becomes feasible again once enough time has passed
        conditions.mjd += 2.0 * visit_gap.gap

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_visit_gap_with_multiple_bands(self):
        visit_gap = basis_functions.VisitGap(note="test", band_names=["g", "i"])

        conditions = Conditions()
        conditions.mjd = 59000.0

        # default is feasible
        assert visit_gap.check_feasibility(conditions=conditions)

        observation = ObservationArray()
        observation["band"] = "r"
        observation["scheduler_note"] = "foo"
        observation["mjd"] = 59000.0

        visit_gap.add_observation(observation=observation)

        # observation with the wrong note
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["scheduler_note"] = "test"
        visit_gap.add_observation(observation=observation)

        # observation with the wrong band
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["band"] = "g"
        observation["mjd"] += 1e-3
        visit_gap.add_observation(observation=observation)

        # observation with the correct note but only one band
        assert visit_gap.check_feasibility(conditions=conditions)

        observation["band"] = "i"
        observation["mjd"] += 1e-3
        visit_gap.add_observation(observation=observation)

        # observation with the correct note and both bands
        assert not visit_gap.check_feasibility(conditions=conditions)

        # make sure it is still not feasible after only the g observation gap
        # has passed
        conditions.mjd += visit_gap.gap + 1.1e-3

        # observation with the correct note and both bands
        assert not visit_gap.check_feasibility(conditions=conditions)

        # make sure it is feasible after both gaps have passed
        conditions.mjd += 1e-3

        assert visit_gap.check_feasibility(conditions=conditions)

    def test_sun_alt(self):
        sunaltbf = basis_functions.SunAltHighLimitBasisFunction(alt_limit=-15)
        conditions = Conditions()
        conditions.sun_alt = np.radians(-20)
        assert not sunaltbf.check_feasibility(conditions)
        conditions.sun_alt = np.radians(-10)
        assert sunaltbf.check_feasibility(conditions)

    def test_close_to_twilight(self):
        bf = basis_functions.CloseToTwilightBasisFunction(
            max_sun_alt_limit=-14.8, max_time_to_12deg=21.0, min_time_remaining=15.0
        )
        conditions = Conditions()
        conditions.mjd = 520900.00
        conditions.sun_alt = np.radians(-14)
        conditions.sun_n12_rising = conditions.mjd + 16.0 / 60 / 24
        assert bf.check_feasibility(conditions)
        conditions.sun_n12_rising = conditions.mjd + 14.0 / 60 / 24
        assert not bf.check_feasibility(conditions)
        conditions.mjd = 520900.00
        conditions.sun_n12_rising = conditions.mjd + 16.0 / 60 / 24
        conditions.sun_alt = np.radians(-20)
        assert bf.check_feasibility(conditions)

    def test_AltAzShadowMask(self):
        nside = 32
        conditions = Conditions(nside=nside)
        conditions.mjd = 59000.0
        conditions.tel_az_limits = [np.radians(-250), np.radians(250)]
        conditions.tel_alt_limits = [np.radians(-100), np.radians(100)]
        # With no (real) limits, including no limits in conditions
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=0, max_alt=90, min_az=0, max_az=360, shadow_minutes=0, pad=0
        )
        result = bf(conditions)
        self.assertTrue(np.all(np.isnan(result[np.where(conditions.alt < 0)])))
        self.assertTrue(np.all(result[np.where(conditions.alt >= 0)] == 0))
        # Set altitude limits but still no padding
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=40, max_alt=60, min_az=0, max_az=360, shadow_minutes=0, pad=0
        )
        result = bf(conditions)
        good = np.where((conditions.alt > np.radians(40)) & (conditions.alt < np.radians(60)), True, False)
        self.assertTrue(np.all(np.isnan(result[~good])))
        self.assertTrue(np.all(result[good] == 0))
        # And set azimuth limits but no padding
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=100, min_az=90, max_az=180, shadow_minutes=0, pad=0
        )
        result = bf(conditions)
        good = np.where((conditions.az > np.radians(90)) & (conditions.az < np.radians(180)), True, False)
        self.assertTrue(np.all(np.isnan(result[~good])))
        self.assertTrue(np.all(result[good] == 0))
        # And set azimuth limits - order sensitive
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=180, max_az=90, shadow_minutes=0, pad=0
        )
        result = bf(conditions)
        good = np.where((conditions.az > np.radians(180)) | (conditions.az < np.radians(90)), True, False)
        self.assertTrue(np.all(np.isnan(result[~good])))
        self.assertTrue(np.all(result[good] == 0))
        # And set altitude limits from the conditions
        conditions.sky_alt_limits = [[np.radians(40), np.radians(60)]]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=0, max_alt=90, min_az=0, max_az=360, shadow_minutes=0
        )
        result = bf(conditions)
        good = np.where((conditions.alt > np.radians(43)) & (conditions.alt < np.radians(57)))
        bad = np.where((conditions.alt < np.radians(42)) & (conditions.alt > np.radians(58)))
        self.assertTrue(np.all(np.isnan(result[bad])))
        self.assertTrue(np.all(result[good] == 0))
        # Set multiple altitude limits from the conditions
        conditions.sky_alt_limits = [[np.radians(40), np.radians(60)], [np.radians(80), np.radians(90)]]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=0, max_az=360, shadow_minutes=0, pad=3
        )
        result = bf(conditions)
        # Conditions value get padded
        good = np.where(
            ((conditions.alt > np.radians(43)) & (conditions.alt < np.radians(57)))
            | ((conditions.alt > np.radians(83)) & (conditions.alt < np.radians(87))),
        )
        bad = np.where(
            (conditions.alt < np.radians(41))
            | ((conditions.alt > np.radians(59)) & (conditions.alt < np.radians(81)))
        )
        self.assertTrue(np.all(np.isnan(result[bad])))
        self.assertTrue(np.all(result[good] == 0))
        # Set azimuth limits
        conditions.sky_alt_limits = None
        conditions.sky_az_limits = [[np.radians(270), np.radians(90)]]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=0, max_az=360, shadow_minutes=0
        )
        result = bf(conditions)
        # Conditions value get padded, but we should be able to mask this area
        # at a minimum
        good = np.where(
            (conditions.az > np.radians(270)) | (conditions.az < np.radians(90)),
            True,
            False,
        )
        self.assertTrue(np.all(np.isnan(result[~good])))
        # Set azimuth limits, direction sensitive
        conditions.sky_alt_limits = None
        conditions.sky_az_limits = [[np.radians(90), np.radians(270)]]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=0, max_az=360, shadow_minutes=0, pad=2
        )
        result = bf(conditions)
        # To more accurately count in azimuth, remove very high altitudes
        good = np.where(
            (conditions.alt < np.radians(80))
            & (conditions.alt > np.radians(0))
            & (conditions.az < np.radians(260))
            & (conditions.az > np.radians(110))
        )
        bad = np.where((conditions.az > np.radians(269)) | (conditions.az < np.radians(91)))
        # Check this area is masked
        self.assertTrue(np.all(np.isnan(result[bad])))
        self.assertTrue(np.all(result[good] == 0))
        # Set altitude limits from kinematic model
        conditions.sky_alt_limits = None
        conditions.tel_alt_limits = [np.radians(20), np.radians(86.5)]
        conditions.sky_az_limits = None
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=0, max_az=360, shadow_minutes=0
        )
        result = bf(conditions)
        good = np.where(
            (conditions.alt > np.radians(23)) & (conditions.alt < np.radians(84)),
        )
        bad = np.where((conditions.alt < np.radians(21)) | (conditions.alt > np.radians(85)))
        # Check this area is masked
        self.assertTrue(np.all(np.isnan(result[bad])))
        self.assertTrue(np.all(result[good] == 0))
        # Set azimuth limits from kinematic model
        conditions.sky_alt_limits = None
        conditions.sky_az_limits = None
        conditions.tel_az_limits = [np.radians(90), np.radians(270)]
        conditions.tel_alt_limits = [np.radians(-100), np.radians(100)]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=-90, max_alt=90, min_az=0, max_az=360, shadow_minutes=0
        )
        result = bf(conditions)
        bad = np.where(
            (conditions.az > np.radians(268)) | (conditions.az < np.radians(92)),
        )
        good = np.where(
            (conditions.alt < np.radians(80))
            & (conditions.alt > np.radians(0))
            & (conditions.az < np.radians(260))
            & (conditions.az > np.radians(110))
        )
        # Check this area is masked
        self.assertTrue(np.all(np.isnan(result[bad])))
        self.assertTrue(np.all(result[good] == 0))
        # Check very simple shadow minutes
        # Set azimuth limits from kinematic model
        conditions.sky_alt_limits = None
        conditions.sky_az_limits = None
        conditions.tel_az_limits = [np.radians(-250), np.radians(250)]
        conditions.tel_alt_limits = [np.radians(-100), np.radians(100)]
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=20, max_alt=85, min_az=0, max_az=360, shadow_minutes=0
        )
        result = bf(conditions)
        bf = basis_functions.AltAzShadowMaskBasisFunction(
            nside=nside, min_alt=20, max_alt=85, min_az=0, max_az=360, shadow_minutes=30
        )
        result_shadow = bf(conditions)
        overlap = np.where(np.isnan(result) & ~np.isnan(result_shadow))
        self.assertTrue(len(overlap[0]) == 0)
        overlap = np.where(np.isnan(result_shadow) & ~np.isnan(result))
        self.assertTrue(len(overlap[0]) > 0)

        # Check that shadow is being computed multiple times
        # basis function where time step is too large, so zenith exclusion path
        # will not be filled properly
        bf_gap = basis_functions.AltAzShadowMaskBasisFunction(shadow_minutes=120.0, time_step=500.0)
        # Better time_step so path of zenith is masked. So bf_filled should
        # have fewer valid pixels.
        bf_filled = basis_functions.AltAzShadowMaskBasisFunction(shadow_minutes=120.0, time_step=10.0)

        mask_gap = bf_gap(conditions)
        mask_filled = bf_filled(conditions)

        assert np.sum(np.isfinite(mask_gap)) > np.sum(np.isfinite(mask_filled))

    def test_deprecated(self):
        # Add to-be-deprecated functions here as they appear
        deprecated_basis_functions = [
            basis_functions.SolarElongMaskBasisFunction,
            basis_functions.AzimuthBasisFunction,
            basis_functions.SeasonCoverageBasisFunction,
            basis_functions.TimeInTwilightBasisFunction,
            basis_functions.FilterLoadedBasisFunction,
            basis_functions.StrictFilterBasisFunction,
            basis_functions.FilterChangeBasisFunction,
            basis_functions.FilterDistBasisFunction,
        ]
        for dep_bf in deprecated_basis_functions:
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always")
                dep_bf()
                # Verify deprecation warning
                assert len(w) >= 1
                assert issubclass(w[-1].category, (DeprecationWarning, FutureWarning))


if __name__ == "__main__":
    unittest.main()



import unittest

import healpy as hp
import numpy as np

from rubin_scheduler.site_models import CloudMap


class TestCloudMap(unittest.TestCase):

    def test_cloudmap(self):
        nside = 32
        n_pix = hp.nside2npix(nside)
        cm = CloudMap()

        n_frames = 10
        rng = np.random.default_rng(seed=42)

        for mjd in np.arange(n_frames):
            cm.add_frame(rng.random(n_pix), mjd)

        map1 = cm.extinction_closest(5)
        assert np.size(map1) == n_pix

        # check that 19 minutes ahead of last frame,
        # still returns closest
        map1 = cm.extinction_closest(mjd + 19 / 60 / 24)
        assert np.size(map1) == n_pix

        # check that going too far from used times
        # returns zero
        map2 = cm.extinction_closest(5000)
        assert map2 == 0

        map3 = cm.extinction_closest(-5000)
        assert map3 == 0

        # Now with uncertainty frames
        cm = CloudMap()
        for mjd in np.arange(n_frames):
            cm.add_frame(rng.random(n_pix), mjd, uncert=rng.random(n_pix))

        # check that 19 minutes ahead of last frame,
        # still returns closest
        map1, uncert1 = cm.extinction_closest(mjd + 19 / 60 / 24, uncert=True)
        assert np.size(map1) == n_pix
        assert np.size(uncert1) == n_pix

        # check that going too far from used times
        # returns zero
        map2, uncert2 = cm.extinction_closest(5000, uncert=True)
        assert map2 == 0
        assert uncert2 == 0


if __name__ == "__main__":
    unittest.main()



__all__ = ("ShortExptDetailer",)

import warnings

import numpy as np

import rubin_scheduler.scheduler.features as features
from rubin_scheduler.scheduler.detailers import BaseDetailer
from rubin_scheduler.scheduler.utils import HpInLsstFov
from rubin_scheduler.utils import DEFAULT_NSIDE, SURVEY_START_MJD


class ShortExptDetailer(BaseDetailer):
    """Check if the area has been observed with a short exposure time
    this year. If not, add some short exposures.

    Parameters
    ----------
    exp_time : `float` (1.)
        The short exposure time to use.
    nobs : `float` (2)
        The number of observations to try and take per year
    night_max : `float` (None)
        Do not apply any changes to the observation list if the current
        night is greater than night_max.
    n_repeat : `int` (1)
        How many short observations to do in a row.
    time_scale : `bool` (False)
        Should the short observations be scaled throughout the year (True),
        or taken as fast as possible (False).

    """

    def __init__(
        self,
        exp_time=1.0,
        bandname="r",
        nside=DEFAULT_NSIDE,
        footprint=None,
        nobs=2,
        mjd0=SURVEY_START_MJD,
        survey_name="short",
        read_approx=2.0,
        night_max=None,
        n_repeat=1,
        time_scale=False,
        filtername=None,
    ):
        if filtername is not None:
            warnings.warn("filtername deprecated in favor of bandname", FutureWarning)
            bandname = filtername
        self.read_approx = read_approx
        self.exp_time = exp_time
        self.bandname = bandname
        self.nside = nside
        self.footprint = footprint
        self.nobs = nobs
        self.survey_name = survey_name
        self.mjd0 = mjd0
        self.night_max = night_max
        self.n_repeat = n_repeat
        self.time_scale = time_scale

        self.survey_features = {}
        # XXX--need a feature that tracks short exposures in the band
        self.survey_features["nobs"] = features.N_observations(
            bandname=bandname, nside=nside, survey_name=self.survey_name
        )
        # Need to be able to look up hpids for each observation
        self.obs2hpid = HpInLsstFov(nside=nside)

    def __call__(self, observation_list, conditions):
        # XXX--this logic would probably make more sense as a feasability
        # basis function. Might consider expanding the detailer base class
        # to include basis functions.
        if self.night_max is not None:
            if conditions.night > self.night_max:
                return observation_list

        out_observations = []
        # Compute how many observations we should have taken by now
        if self.time_scale:
            n_goal = self.nobs * np.ceil(conditions.night / self.night_max)
        else:
            n_goal = self.nobs
        time_to_add = 0.0
        for observation in observation_list:
            out_observations.append(observation)
            if observation["band"] == self.bandname:
                hpids = self.obs2hpid(observation["RA"], observation["dec"])
                # Crop off anything outside the target footprint
                hpids = hpids[np.where(self.footprint[hpids] > 0)]
                # Crop off things where we already have enough observation
                hpids = hpids[np.where(self.survey_features["nobs"].feature[hpids] < n_goal)]
                if np.size(hpids) > 0:
                    for i in range(0, self.n_repeat):
                        new_obs = observation.copy()
                        new_obs["exptime"] = self.exp_time
                        new_obs["nexp"] = 1
                        new_obs["scheduler_note"] = self.survey_name
                        out_observations.append(new_obs)
                        time_to_add += new_obs["exptime"] + self.read_approx
        # pump up the flush time
        for observation in observation_list:
            observation["flush_by_mjd"] += time_to_add / 3600.0 / 24.0
        return out_observations



import unittest

from astropy.coordinates import SkyCoord

from rubin_scheduler.utils import (
    ddf_locations,
    ddf_locations_pre3_5,
    special_locations,
)


class LocationsTests(unittest.TestCase):

    def test_ddf(self):
        old_locations = ddf_locations_pre3_5()

        for key in old_locations:
            assert isinstance(old_locations[key][0], float)
            assert isinstance(old_locations[key][1], float)

        ddf_loc = ddf_locations()
        for key in ddf_loc:
            assert isinstance(ddf_loc[key][0], float)
            assert isinstance(ddf_loc[key][1], float)

        ddf_skycoords = ddf_locations(skycoords=True)
        for key in ddf_skycoords:
            assert isinstance(ddf_skycoords[key], SkyCoord)

    def test_locations(self):

        locations = special_locations()
        for key in locations:
            assert isinstance(locations[key][0], float)
            assert isinstance(locations[key][1], float)

        locations = special_locations(skycoords=True)
        for key in locations:
            assert isinstance(locations[key], SkyCoord)


if __name__ == "__main__":
    unittest.main()



__all__ = ("CloudData", "ConstantCloudData")

import os
import sqlite3
from dataclasses import dataclass

import numpy as np
from astropy.time import Time

from rubin_scheduler.data import get_data_dir


class CloudData:
    """Handle the cloud information.

    This class deals with the cloud information that was previously
    produced for OpSim version 3.

    Parameters
    ----------
    start_time : astropy.time.Time
        The time of the start of the simulation.
        The cloud database will be assumed to start on Jan 01 of
        the same year.
    cloud_db : str, optional
        The full path name for the cloud database. Default None,
        which will use the database stored in the module
        (site_models/clouds_ctio_1975_2022.db).
    offset_year : float, optional
        Offset into the cloud database by 'offset_year' years. Default 0.
    scale : float (1e6)
        Enforce machine precision for cross-platform repeatability by
        scaling and rounding date values.
    """

    def __init__(self, start_time, cloud_db=None, offset_year=0, scale=1e6):
        self.cloud_db = cloud_db
        if self.cloud_db is None:
            self.cloud_db = os.path.join(get_data_dir(), "site_models", "clouds_ctio_1975_2022.db")

        # Cloud database starts in Jan 01 of the year of the
        # start of the simulation.
        year_start = start_time.datetime.year + offset_year
        self.start_time = Time("%d-01-01" % year_start, format="isot", scale="tai")

        self.cloud_dates = None
        self.cloud_values = None
        self.scale = scale
        self.read_data()

    def __call__(self, time):
        """Get the cloud for the specified time.

        Parameters
        ----------
        time : `astropy.time.Time`
            Time in the simulation for which to find the current cloud
            coverage. The difference between this time and the start_time,
            plus the offset, will be used to query the cloud database
            for the 'current' conditions.

        Returns
        -------
        cloud_value : `float`
            The fraction of the sky that is cloudy (measured in
            steps of 8ths) closest to the specified time.
        """
        delta_time = (time - self.start_time).sec
        dbdate = delta_time % self.time_range + self.min_time
        if self.scale is not None:
            dbdate = np.round(dbdate * self.scale).astype(int)
        idx = np.searchsorted(self.cloud_dates, dbdate)
        # searchsorted ensures that left < date < right
        # but we need to know if date is closer to left or to right
        left = self.cloud_dates[idx - 1]
        right = self.cloud_dates[idx]
        # If we are only doing one time
        if np.size(left) == 1:
            if dbdate - left < right - dbdate:
                idx -= 1
        # If we have an array of times
        else:
            d1 = dbdate - left
            d2 = right - dbdate
            to_sub = np.where(d1 < d2)
            idx[to_sub] -= 1

        return self.cloud_values[idx]

    def read_data(self):
        """Read the cloud data from disk.

        The default behavior is to use the module stored database.
        However, an alternate database file can be provided. The alternate
        database file needs to have a table called *Cloud* with the
        following columns:

        cloudId
            int : A unique index for each cloud entry.
        c_date
            int : The time (units=seconds) since the start of the
            simulation for the cloud observation.
        cloud
            float : The cloud coverage (in steps of 8ths) of the sky.
        """
        with sqlite3.connect(self.cloud_db) as conn:
            cur = conn.cursor()
            query = "select c_date, cloud from Cloud order by c_date;"
            cur.execute(query)
            results = np.array(cur.fetchall())
            self.cloud_dates = np.hsplit(results, 2)[0].flatten()
            self.cloud_values = np.hsplit(results, 2)[1].flatten()
            cur.close()
        # Make sure seeing dates are ordered appropriately (monotonically
        # increasing).
        ordidx = self.cloud_dates.argsort()
        self.cloud_dates = self.cloud_dates[ordidx]
        # Record this information, in case the cloud database does not
        # start at t=0.
        self.min_time = self.cloud_dates[0]
        self.max_time = self.cloud_dates[-1]
        self.time_range = self.max_time - self.min_time
        if self.scale is not None:
            self.cloud_dates = np.round(self.cloud_dates * self.scale).astype(int)
        self.cloud_values = self.cloud_values[ordidx]


@dataclass
class ConstantCloudData:
    """Generate constant cloud information.

    Parameters
    ----------
    cloud_fraction : `float`
        The fraction of the sky that is cloudy.
    """

    cloud_fraction: float = 0.0

    def __call__(self, time):
        """Get the cloud for the specified time.

        Parameters
        ----------
        time : `astropy.time.Time`
            In principle, the time in the simulation for which to find the
            current cloud coverage. In practice, this argument is ignored.

        Returns
        -------
        cloud_faction : `float`
            The fraction of the sky that is cloudy.
        """
        return self.cloud_fraction
